
//===-- temu-c/Memory.h - Memory Helpers ------------------------*- C++ -*-===//
//
// T-EMU: The Terma Emulator
// (c) Terma 2015-2016
// Authors: Mattias Holm <maho (at) terma.com>
//
//===----------------------------------------------------------------------===//


#ifndef TEMU_SUPPORT_MEMORY
#define TEMU_SUPPORT_MEMORY

#include <stdint.h>
#include "temu-c/Support/Attributes.h"
#include "temu-c/Support/Objsys.h"

#ifdef __cplusplus
extern "C" {
#endif

  //! Memory attribute enumeration

  //! The emulator provides 5 standard attributes, and 3 user defined
  //! ones. The attributes are set in the memory space (not the memory
  //! models), so it is possible to set a watch point on memory mapped
  //! devices. When an attribute is set on a page, that page will get a
  //! shadow attribute page (same size as the page), enabling attributes
  //! to be set on a per byte level.
  //!
  //! Attributes are only checked on the address being accessed, the
  //! transaction size is not taken into account.
  typedef enum {
    teMA_Break      = 1,      //!< Breakpoint set
    teMA_WatchRead  = 1 << 1, //!< Read watchpoint set
    teMA_WatchWrite = 1 << 2, //!< Write watchpoint set
    teMA_Upset      = 1 << 3, //!< Single event upset
    teMA_Faulty     = 1 << 4, //!< Multiple event upset / uncorrectable
    teMA_User1      = 1 << 5, //!< User definable
    teMA_User2      = 1 << 6, //!< User definable
    teMA_User3      = 1 << 7, //!< User definable
  } temu_MemoryAttr;

typedef uint8_t temu_MemoryAttrs;


// Private interface, this is not stable yet
typedef struct temu_MemorySpaceIface {
  int (*mapDevice)(void *Obj, uint64_t Addr, uint64_t Len, void *Device,
                   uint32_t Flags);
  void (*setAttr)(void *Obj, uint64_t Addr, uint64_t Len, temu_MemoryAttr Attr);
  void (*clearAttr)(void *Obj, uint64_t Addr, uint64_t Len, temu_MemoryAttr Attr);
  temu_MemoryAttrs (*getAttrs)(void *Obj, uint64_t Addr);

  int (*mapDeviceWithNamedIface)(void *Obj, uint64_t Addr, uint64_t Len, void *Device, const char *IfaceName,
                                 uint32_t Flags);
  int (*unmapRange)(void *Obj, uint64_t Addr, uint64_t Len);
} temu_MemorySpaceIface;

TEMU_IFACE_REFERENCE_TYPE(temu_MemorySpace);
#define TEMU_MEMORY_SPACE_IFACE_TYPE "MemorySpaceIface"

// Private interface, do not use directly
typedef struct {
  void* (*getPDC)(void *Obj, uint64_t Addr);
  void (*allocPDC)(void *Obj, uint64_t Addr);
  void (*disposePDC)(void *Obj, uint64_t Addr);
  void (*writePDC)(void *Obj, uint64_t Addr, void *Inst);
  void (*clearPDC)(void *Obj, uint64_t Addr);
} temu_PDCIface;
TEMU_IFACE_REFERENCE_TYPE(temu_PDC);
#define TEMU_PDC_IFACE_TYPE "PDCIface"


/*!
 * temu_memoryRead TODO
 * \param mem Pointer to memory space object
 * \param buff The buffer to which the memory should be stored
 * \param addr The address of the memory block to be read
 * \param size The size to be read
 * \param swap setting this to 0 indicates reading a byte array, 1 a
 *   uint16 array, 2 a uint32 array and 3 a uint64 array
 *
 * \result Negative on failures. Other values indicate success. The
 *   convention is to return the number of bytes read which should be
 *   == size.
 */
TEMU_API int temu_memoryRead(void *mem, uint8_t *buff, uint64_t addr, uint32_t size,
                    int swap);

/*!
 * temu_memoryWrite TODO
 * \param mem Pointer to memory space object
 * \param buff The buffer to which the memory should be stored
 * \param addr The address, at which the write should be done
 * \param size The size to be read
 * \param swap Negative on failures. Other values indicate success.
 *
 * \result Negative on failure. Other values indicate success. The
 *   convention is to return the number of bytes written, which should
 *   be == size.
 */
TEMU_API int temu_memoryWrite(void *mem, uint64_t addr, uint8_t *buff, uint32_t size,
                              int swap);

/*!
 * temu_memoryMap Maps an object into a memory space. The object must have an interface named MemAccessIface, of the type defined as TEMU_MEM_ACCESS_IFACE_TYPE
 * \param Obj is the memory space object
 * \param Addr the physical address
 * \param Len the length in bytes
 * \param MemObj is the object that is mapped into the memory space it needs to be of a class that follows the rules above
 * \param Flags are flags that are copied into the memory transaction object's flag field when a transaction reaches an object
 * \result Zero on success
 */
TEMU_API int temu_memoryMap(void *Obj, uint64_t Addr, uint64_t Len, void *MemObj,
                            uint32_t Flags);


/*!
 * temu_memoryMapNamedIface Maps an object into a memory space using the named memory access interface.
 * The interface named by IfaceName must be of the type TEMU_MEM_ACCESS_IFACE_TYPE
 * \param Obj is the memory space object
 * \param Addr the physical address
 * \param Len the length in bytes
 * \param MemObj is the object that is mapped into the memory space it needs to be of a class that follows the rules above
 * \param IfaceName Name of the memory access interface
 * \param Flags are flags that are copied into the memory transaction object's flag field when a transaction reaches an object
 * \result Zero on success
 */
TEMU_API int temu_memoryMapNamedIface(void *Obj, uint64_t Addr, uint64_t Len, void *MemObj, const char *IfaceName,
                                      uint32_t Flags);

/*!
 * Sets an attribute on the given memory range
 * \param Obj  Memory space object
 * \param Addr Physical address of start
 * \param Len Length of memory range
 * \param Attr Attribute to set
 */
TEMU_API void temu_memorySetAttr(void *Obj, uint64_t Addr, uint64_t Len, temu_MemoryAttr Attr);

/*!
 * Clears an attribute on the given memory range
 * \param Obj  Memory space object
 * \param Addr Physical address of start
 * \param Len Length of memory range
 * \param Attr Attribute to clear
 */
TEMU_API void temu_memoryClearAttr(void *Obj, uint64_t Addr, uint64_t Len, temu_MemoryAttr Attr);

/*!
 * \brief Get memory attributes for address
 * \param Obj Memory space
 * \param Addr Physical addres
 * \result Memory attributes for the address
 */
TEMU_API temu_MemoryAttrs temu_memoryGetAttrs(void *Obj, uint64_t Addr);

/*!
 * Issue a memory read transaction (without initiator)
 * \param Obj Memory space
 * \param Addr Physical address
 * \param Word 32 bit word that is read
 * \result 0 on success, other values imply failure and will not
 *         update the word.
 */
TEMU_API int temu_memoryReadPhys32(void *Obj, uint64_t Addr, uint32_t *Word);

/*!
 * Issue a memory write transaction (without initiator)
 * \param Obj Memory space
 * \param Addr Physical address
 * \param Word 32 bit word to write
 * \result 0 on success
 */
TEMU_API int temu_memoryWritePhys32(void *Obj, uint64_t Addr, uint32_t Word);

/*!
 * Normalise a value for writes where only 32 bit transactions are
 * supported by the device model, but the target is allowed to write
 * non-32 bit quantity. The function mixes the old register value with
 * the new one based on the size and offset parameter. The problem
 * exists because the memory access interface has a value entry, which
 * always ends up in the lower bits, so if we write to the higher bits
 * in in a 32 bit register, then by just forwarding the the value as
 * is to the write handler, will result in a write of the lower bits and
 * a clear of the upper bits (for a store unsigned).
 *
 * Thus a normalisation is neded where we mix the written word with the old
 * bits. So for transaction size of 16, and an offset of 16, the resulting
 * word is (old & 0x0000ffff) | (new << 16)
 *
 * \param OldVal Old regiseter value.
 * \param NewVal Content in memory transaction (new value).
 * \param Sz Size in log number of bytes
 * \param Off Offset in bytes within the 32 bit word of the transaction.
 */
TEMU_API uint32_t temu_normaliseWrite32(uint32_t OldVal, uint32_t NewVal,
                                        int Sz, int Off);

/*!
 * Normalise a value for reads where only 32 bit transactions are
 * supported by the device model, but the target is allowed to read
 * non-32 bit quantities.
 *
 * Given a 32 bit register value, the function extracts the bits from it
 * that was actually requested.
 *
 * \param Value Register value
 * \param Sz Size in log number of bytes (0, 1, or 2)
 * \param Off Offset in bytes within the 32 bit word of the transaction (0-3)
 */
TEMU_API uint32_t temu_normaliseRead32(uint32_t Value, int Sz, int Off);


/*!
 * Normalise a value for reads where only 16 bit transactions are
 * supported by the device model, but the target is allowed to read
 * non-16 bit quantities.
 *
 * Given a 32 bit register value, the function extracts the bits from it
 * that was actually requested.
 *
 * \param Value Register value
 * \param Sz Size in log number of bytes (0 or 1)
 * \param Off Offset in bytes within the 32 bit word of the transaction (0 or 2)
 */
TEMU_API uint16_t temu_normaliseRead16(uint16_t Value, int Sz, int Off);


/*!
 * Normalise a value for writes where only 16 bit transactions are
 * supported by the device model, but the target is allowed to write
 * non-16 bit quantity. The function mixes the old register value with
 * the new one based on the size and offset parameter. The problem
 * exists because the memory access interface has a value entry, which
 * always ends up in the lower bits, so if we write to the higher bits
 * in in a 16 bit register, then by just forwarding the the value as
 * is to the write handler, will result in a write of the lower bits and
 * a clear of the upper bits (for a store unsigned).
 *
 * Thus a normalisation is neded where we mix the written word with the old
 * bits. So for transaction size of 8, and an offset of8, the resulting
 * word is (old & 0x00ff) | (new << 8)
 *
 * \param OldVal Old regiseter value.
 * \param NewVal Content in memory transaction (new value).
 * \param Sz Size in log number of bytes
 * \param Off Offset in bytes within the 16 bit word of the transaction.
 */

TEMU_API uint16_t temu_normaliseWrite16(uint16_t OldVal, uint16_t NewVal,
                                        int Sz, int Off);



#ifdef __cplusplus
}
#endif

#endif /* ! TEMU_SUPPORT_MEMORY */
